//############################################################################
//
// LaserBoy !!!
//
// by James Lehman
// Extra Stimulus Inc.
// james@akrobiz.com
//
// began: October 2003
//
// Copyright 2003, 2004, 2005, 2006, 2007, 2008, 2009 James Lehman.
// This source is distributed under the terms of the GNU General Public License.
//
// LaserBoy_frame_set.cpp is part of LaserBoy.
//
// LaserBoy is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// LaserBoy is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with LaserBoy. If not, see <http://www.gnu.org/licenses/>.
//
//############################################################################
#include "LaserBoy_GUI_base.hpp"

//############################################################################
LaserBoy_frame_set::LaserBoy_frame_set(LaserBoy_space* ps,
                                       LaserBoy_vertex(*F)(int, int),
                                       int vertices_per_frame,
                                       int num_frames
                                      )
                   : p_space            (ps)
                   , frame_select_start (-1)
                   , frame_select_end   (-1)
                   , error              (LASERBOY_OK)
                   , null_frame         (ps)
                   , intro              (ps)
{
    null_frame = NULL_frame(ps);
    reserve(num_frames);
    for(frame_index = 0; frame_index < num_frames; frame_index++)
        push_back(LaserBoy_frame(p_space, F, vertices_per_frame, frame_index));
    renumber();
    frame_index = 0;
    error = LASERBOY_OK;
}

//############################################################################
LaserBoy_frame_set& LaserBoy_frame_set::operator += (const LaserBoy_frame& frame)
{
    if(p_space->show_effects_generation)
    {
        frame.render();
        p_space->GUI->display_space();
    }
    push_back(frame);
    renumber();
    return *this;
}

//############################################################################
bool LaserBoy_frame_set::from_ild_file(string file, bool append)
{
    ifstream in(file.c_str(), ios::in | ios::binary);

    if(in.is_open())
    {
        frame_index = 0;
        error       = LASERBOY_OK;
        error       = from_ifstream_ild(in, append);
        in.close();
    }
    else
        error = LASERBOY_FILE_OPEN_FAILED;

    if(error)
        return false;
    return true;
}

//############################################################################
bool LaserBoy_frame_set::from_txt_file(string file, bool append)
{
    ifstream in(file.c_str(), ios::in); // not binary!

    if(in.is_open())
    {
        frame_index = 0;
        error       = LASERBOY_OK;
        error       = from_ifstream_txt(in, append);
        in.close();
    }
    else
        error = LASERBOY_FILE_OPEN_FAILED;

    if(error)
        return false;
    return true;
}

//############################################################################
bool LaserBoy_frame_set::superimpose_from_ild_file(string file)
{
    int                 i;
    LaserBoy_frame_set  overlay(p_space),
                        sum    (p_space);

    if(    overlay.from_ild_file(file)
        && overlay.number_of_frames()
      )
    {
        if(overlay.number_of_frames() > 1 && number_of_frames() > 1)
        {
            if(number_of_frames() > overlay.number_of_frames())
            {
                for(i = 0; i < overlay.number_of_frames(); i++)
                    sum += (frame(i) + overlay.frame(i));

                for(i = overlay.number_of_frames(); i < number_of_frames(); i++)
                    sum += frame(i);
            }
            else if(number_of_frames() < overlay.number_of_frames())
            {
                for(i = 0; i < number_of_frames(); i++)
                    sum += (frame(i) + overlay.frame(i));

                for(i = number_of_frames(); i < overlay.number_of_frames(); i++)
                    sum += overlay.frame(i);
            }
            else
            {
                for(i = 0; i < number_of_frames(); i++)
                    sum += (frame(i) + overlay.frame(i));
            }
        }
        else if(overlay.number_of_frames() == 1)
        {
            for(i = 0; i < number_of_frames(); i++)
                sum += (frame(i) + overlay.frame(0));
        }
        else if(number_of_frames() == 1)
        {
            for(i = 0; i < overlay.number_of_frames(); i++)
                sum += (frame(0) + overlay.frame(i));
        }
        *this = sum;
        return true;
    }
    return false;
}

//############################################################################
bool LaserBoy_frame_set::superimpose_from_txt_file(string file)
{
    int                 i;
    LaserBoy_frame_set  overlay(p_space),
                        sum    (p_space);

    if(    overlay.from_txt_file(file)
        && overlay.number_of_frames()
      )
    {
        if(overlay.number_of_frames() > 1 && number_of_frames() > 1)
        {
            if(number_of_frames() > overlay.number_of_frames())
            {
                for(i = 0; i < overlay.number_of_frames(); i++)
                    sum += (frame(i) + overlay.frame(i));

                for(i = overlay.number_of_frames(); i < number_of_frames(); i++)
                    sum += frame(i);
            }
            else if(number_of_frames() < overlay.number_of_frames())
            {
                for(i = 0; i < number_of_frames(); i++)
                    sum += (frame(i) + overlay.frame(i));

                for(i = number_of_frames(); i < overlay.number_of_frames(); i++)
                    sum += overlay.frame(i);
            }
            else
            {
                for(i = 0; i < number_of_frames(); i++)
                    sum += (frame(i) + overlay.frame(i));
            }
        }
        else if(overlay.number_of_frames() == 1)
        {
            for(i = 0; i < number_of_frames(); i++)
                sum += (frame(i) + overlay.frame(0));
        }
        else if(number_of_frames() == 1)
        {
            for(i = 0; i < overlay.number_of_frames(); i++)
                sum += (frame(0) + overlay.frame(i));
        }
        *this = sum;
        return true;
    }
    return false;
}

//############################################################################
bool LaserBoy_frame_set::from_dxf_file(string file, bool append)
{
    ifstream in(file.c_str(), ios::in);

    if(!append)
    {
        clear();
        num_2D_frames = 0;
        num_3D_frames = 0;
        frame_index   = 0;
    }

    if(in.is_open())
    {
        LaserBoy_frame frame(p_space);
        error = LASERBOY_OK;
        if(LASERBOY_OK == (error = frame.from_ifstream_dxf(in)))
        {
            p_space->palette_index = frame.palette_index;
            push_back(frame);
            renumber();
        }
        in.close();
    }
    else
        error = LASERBOY_FILE_OPEN_FAILED;

    if(error)
        return false;

    if(at(0).find_rgb_in_palette(p_space->palette_picker(LASERBOY_DXF)))
        at(0).palette_index = LASERBOY_DXF;
        
    return true;
}

//############################################################################
bool LaserBoy_frame_set::from_dxf_directory(string dir, bool append)
{
    DIR                    *pDIR;
    struct dirent          *pDirEnt;
    vector<string>          file_list;
    //------------------------------------------------------------------------
    pDIR = opendir(dir.c_str());
    //------------------------------------------------------------------------
    if(pDIR == NULL)
        error = LASERBOY_FILE_OPEN_FAILED;
    else
    {
        pDirEnt = readdir(pDIR);
        while(pDirEnt != NULL)
        {
            if(    strcmp(pDirEnt->d_name, "." ) // not the dot
                && strcmp(pDirEnt->d_name, "..") // not the dot dot
              )
                file_list.push_back(pDirEnt->d_name);
            pDirEnt = readdir(pDIR);
        }
        closedir(pDIR);
        if(file_list.size())
        {
            bool  valid_data = false;
            int   i, j;

            LaserBoy_real_segment      real_vertices(p_space);
            LaserBoy_real_segment_set  real_segments(p_space);
            LaserBoy_vertex            vertex;
            LaserBoy_frame             frame(p_space);
            //----------------------------------------------------------------
            frame.format        = LASERBOY_3D_FRAME ;
            frame.palette_index = LASERBOY_TRUE_COLOR;
            //----------------------------------------------------------------
            if(!append)
            {
                clear();
                num_2D_frames = 0;
                num_3D_frames = 0;
            }
            //----------------------------------------------------------------
            for(i = 0; i < (int)file_list.size(); i++)
            {
                ifstream in((dir + '/' + file_list[i]).c_str(), ios::in);
                if(in.is_open())
                {
                    real_vertices.from_ifstream_dxf(in);
                    if(real_vertices.size() > 2) // there is a picture in there
                        valid_data = true;
                    real_segments.push_back(real_vertices);
                    real_vertices.clear();
                    in.close();
                }
            }
            //----------------------------------------------------------------
            if(valid_data)
            {
                real_segments.normalize();
                reserve(real_segments.size());
                //------------------------------------------------------------
                for(j = 0; j < (int)real_segments.size(); j++)
                {
                    if(real_segments[j].size() > 2)
                    {
                        for(i = 2; i < (int)real_segments[j].size(); i++)
                        {
                            vertex = (LaserBoy_vertex)real_segments[j][i];
                            if(real_segments[j][i].is_blank())
                                vertex.blank();
                            else
                                vertex.unblank();
                            vertex.set_unique();
                            frame.push_back(vertex);
                            vertex.clear();
                        }
                    } // end if(real_segments[j].size() > 2)
                    else
                    {
                        frame.push_back(vertex);
                        frame.push_back(vertex);
                    }
                    if(frame.find_rgb_in_palette(p_space->palette_picker(LASERBOY_DXF)))
                        frame.palette_index = LASERBOY_DXF;
                    frame.is_2D();
                    push_back(frame);
                    frame.empty(); // clears frame and sets quantity to zero!
                }
                //------------------------------------------------------------
                renumber();
            } // end if(valid_data)
            else
            {
                frame.push_back(vertex);
                frame.push_back(vertex);
                frame.palette_index = LASERBOY_DXF;
                push_back(frame);
            }
        //--------------------------------------------------------------------
        } // end if(file_list.size())
        else
            error = LASERBOY_FILE_OPEN_FAILED;
    }
    if(error)
        return false;
    return true;
}

//############################################################################
bool LaserBoy_frame_set::from_LaserBoy_wave_file(string file, bool append)
{
    ifstream in(file.c_str(), ios::in | ios::binary);
    if(in.is_open())
    {
        LaserBoy_wave_header header(in);
        frame_index = 0;
        //--------------------------------------------------------------------
        if(header.version == "!LaserBoy!")
            error = from_ifstream_unformatted_wave(header, in, append);
        //--------------------------------------------------------------------
        else if(header.LaserBoy_wave_mode & LASERBOY_WAVE_END_OF_FRAME)
            error = from_ifstream_LaserBoy_wave(header, in, append);
        //--------------------------------------------------------------------
        else
            error = from_ifstream_unframed_wave(header, in, append);
        //--------------------------------------------------------------------
        in.close();
    }
    else
        error = LASERBOY_FILE_OPEN_FAILED;

    if(error)
        return false;
    return true;
}

//############################################################################
bool LaserBoy_frame_set::from_unframed_wave_file(string file, bool append)
{
    ifstream in(file.c_str(), ios::in | ios::binary);
    if(in.is_open())
    {
        LaserBoy_wave_header header(in);
        frame_index = 0;
        error       = from_ifstream_unframed_wave(header, in, append);
        in.close();
    }
    else
        error = LASERBOY_FILE_OPEN_FAILED;

    if(error)
        return false;
    return true;
}

//############################################################################
int LaserBoy_frame_set::from_ifstream_ild(ifstream& in, bool append)
{
    LaserBoy_header  header                    ;
    LaserBoy_frame   frame   (p_space)         ;
    LaserBoy_palette palette (p_space)         ;
    bool             true_color_data    = false,
                     some_data_accepted = false;
    //------------------------------------------------------------------------
    if(!append)
    {
        clear();
        num_2D_frames = 0;
        num_3D_frames = 0;
    }
    //------------------------------------------------------------------------
    p_space->palette_index = LASERBOY_ILDA_DEFAULT;
    //------------------------------------------------------------------------
    while(header.from_ifstream_ild(in))
    {
        if(header.quantity)
        {
            switch(header.format)
            {
                //------------------------------------------------------------
                case LASERBOY_3D_FRAME:
                case LASERBOY_2D_FRAME:
                    error |= frame.from_ifstream_ild(in, header);
                    if(!error)
                    {
                        frame.palette_index = p_space->palette_index;
                        frame.set_rgb_from_palette();
                        frame.is_2D() ? num_2D_frames++ : num_3D_frames++;
                        push_back(frame);
                        some_data_accepted = true;
                    }
                    break;
                //------------------------------------------------------------
                case LASERBOY_PALETTE:
                    error |= palette.from_ifstream_ild(in, header);
                    if(!error)
                    {
                        bool built_in_palette = false;
                        for(int i = 0; i < LASERBOY_BUILT_IN_PALETTES; i++)
                            if(header.name == p_space->palette_picker(i).name)
                            {
                                p_space->palette_index = i;
                                built_in_palette = true;
                                break;
                            }
                        if(!built_in_palette)
                        {
                            p_space->push_back_palette(palette);
                            p_space->palette_index = p_space->number_of_palettes() - 1;
                        }
                        some_data_accepted = true;
                    }
                    break;
                //------------------------------------------------------------
                case LASERBOY_TABLE:
                    error |= frame.from_ifstream_format_3(in, header);
                    if(!error)
                    {
                        true_color_data = true;
                        frame.is_2D() ? num_2D_frames++ : num_3D_frames++;
                        push_back(frame);
                        some_data_accepted = true;
                    }
                    break;
                //------------------------------------------------------------
                case LASERBOY_3D_FRAME_RGB:
                case LASERBOY_2D_FRAME_RGB:
                    error |= frame.from_ifstream_ild(in, header);
                    if(!error)
                    {
                        frame.palette_index = LASERBOY_TRUE_COLOR;
                        frame.is_2D() ? num_2D_frames++ : num_3D_frames++;
                        push_back(frame);
                        some_data_accepted = true;
                    }
                    break;
                //------------------------------------------------------------
            }
        }
    }
    //------------------------------------------------------------------------
    if(!in.eof() && error && some_data_accepted)
        error = LASERBOY_OK;
    //------------------------------------------------------------------------
    if(true_color_data)
        p_space->palette_index = LASERBOY_TRUE_COLOR;
    //------------------------------------------------------------------------
    return error;
}
//############################################################################
int LaserBoy_frame_set::from_ifstream_txt(ifstream& in, bool append)
{
    int               i,
                      character,
                      frame_count = 0,
                      line_number = 1,
                      named_palette_index;
                      
    vector<int>       real_frames_index;

    string            word;
    
    LaserBoy_frame    frame   (p_space);
    LaserBoy_palette  palette (p_space);

    LaserBoy_real_segment      real_segment (p_space);
    LaserBoy_real_segment_set  real_segments(p_space);
    //------------------------------------------------------------------------
    if(!append)
    {
        clear();
        num_2D_frames = 0;
        num_3D_frames = 0;
    }
    //------------------------------------------------------------------------
    while(in.good())
    {
        clear_to_alpha(in, line_number);
        if(get_next_word(in, word, line_number))
        {
            //----------------------------------------------------------------
            if(word == "frame")
            {
                if(get_next_word(in, word, line_number))
                {
                    if(word == "xy") // 2D
                    {
                        // frame xy
                        if(get_next_word(in, word, line_number))
                        {
                            if(word == "rgb")
                            {
                                // frame xy rgb
                                if(get_next_word(in, word, line_number))
                                {
                                    if(word == "short")
                                    {
                                        // frame xy rgb short
                                        frame.from_ifstream_txt(in, LASERBOY_2D_FRAME_RGB, LASERBOY_SHORT, line_number);
                                        push_back(frame);
                                        frame_count++;
                                    }
                                    else if(word == "unit")
                                    {
                                        // frame xy rgb unit
                                        frame.from_ifstream_txt(in, LASERBOY_2D_FRAME_RGB, LASERBOY_UNIT, line_number);
                                        push_back(frame);
                                        frame_count++;
                                    }
                                    else if(word == "real")
                                    {
                                        // frame xy rgb real
                                        real_segment.from_ifstream_txt(in, LASERBOY_2D_FRAME_RGB, line_number);
                                        real_segments.push_back(real_segment);
                                        real_frames_index.push_back(frame_count);
                                        frame.empty(); // clears frame and sets quantity to zero!
                                        push_back(frame);
                                        frame_count++;
                                    }
                                    else
                                        // frame xy rgb bad_word!
                                        cout << "error on line: " << line_number << " \'" << word << "\' is not a LaserBoy text format token." << endl;
                                }
                                else
                                {
                                    // frame xy rgb short
                                    frame.from_ifstream_txt(in, LASERBOY_2D_FRAME_RGB, LASERBOY_SHORT, line_number);
                                    push_back(frame);
                                    frame_count++;
                                }
                            }
                            else if(word == "hex")
                            {
                                // frame xy hex
                                if(get_next_word(in, word, line_number))
                                {
                                    if(word == "short")
                                    {
                                        // frame xy hex short
                                        frame.from_ifstream_txt(in, LASERBOY_2D_FRAME_HEX, LASERBOY_SHORT, line_number);
                                        push_back(frame);
                                        frame_count++;
                                    }
                                    else if(word == "unit")
                                    {
                                        // frame xy hex unit
                                        frame.from_ifstream_txt(in, LASERBOY_2D_FRAME_HEX, LASERBOY_UNIT, line_number);
                                        push_back(frame);
                                        frame_count++;
                                    }
                                    else if(word == "real")
                                    {
                                        // frame xy hex real
                                        real_segment.from_ifstream_txt(in, LASERBOY_2D_FRAME_HEX, line_number);
                                        real_segments.push_back(real_segment);
                                        real_frames_index.push_back(frame_count);
                                        frame.empty();
                                        push_back(frame);
                                        frame_count++;
                                    }
                                    else
                                        // frame xy hex bad_word!
                                        cout << "error on line: " << line_number << " \'" << word << "\' is not a LaserBoy text format token." << endl;
                                }
                                else
                                {
                                    // frame xy hex short
                                    frame.from_ifstream_txt(in, LASERBOY_2D_FRAME_HEX, LASERBOY_SHORT, line_number);
                                    push_back(frame);
                                    frame_count++;
                                }
                            }
                            else if(word == "palette")
                            {
                                // frame xy palette
                                if(get_next_word(in, word, line_number))
                                {
                                    if(word == "short")
                                    {
                                        // frame xy palette short
                                        frame.from_ifstream_txt(in, LASERBOY_2D_FRAME_PALETTE, LASERBOY_SHORT, line_number);
                                        push_back(frame);
                                        frame_count++;
                                    }
                                    else if(word == "unit")
                                    {
                                        // frame xy palette unit
                                        frame.from_ifstream_txt(in, LASERBOY_2D_FRAME_PALETTE, LASERBOY_UNIT, line_number);
                                        push_back(frame);
                                        frame_count++;
                                    }
                                    else if(word == "real")
                                    {
                                        // frame xy palette real
                                        real_segment.from_ifstream_txt(in, LASERBOY_2D_FRAME_PALETTE, line_number);
                                        real_segments.push_back(real_segment);
                                        real_frames_index.push_back(frame_count);
                                        frame.empty();
                                        push_back(frame);
                                        frame_count++;
                                    }
                                    else
                                        // frame xy palette bad_word!
                                        cout << "error on line: " << line_number << " \'" << word << "\' is not a LaserBoy text format token." << endl;
                                }
                                else
                                {
                                    // frame xy palette short
                                    frame.from_ifstream_txt(in, LASERBOY_2D_FRAME_PALETTE, LASERBOY_SHORT, line_number);
                                    push_back(frame);
                                    frame_count++;
                                }
                            }
                            else if(word == "table")
                            {
                                // frame xy table
                                if(get_next_word(in, word, line_number))
                                {
                                    if(word == "short")
                                    {
                                        // frame xy table short
                                        if(number_of_frames())
                                        {
                                            back().from_ifstream_txt(in, LASERBOY_2D_FRAME_TABLE, LASERBOY_SHORT, line_number);
                                            frame_count++;
                                        }
                                    }
                                    else if(word == "unit")
                                    {
                                        // frame xy table unit
                                        if(number_of_frames())
                                        {
                                            back().from_ifstream_txt(in, LASERBOY_2D_FRAME_TABLE, LASERBOY_UNIT, line_number);
                                            frame_count++;
                                        }
                                    }
                                    else if(word == "real")
                                    {
                                        // frame xy table real
                                        real_segment.from_ifstream_txt(in, LASERBOY_2D_FRAME_TABLE, line_number);
                                        for(int i = 0; i < back().number_of_vertices(); i++)
                                        {
                                            real_segment[i + 2].r = back()[i].r;
                                            real_segment[i + 2].g = back()[i].g;
                                            real_segment[i + 2].b = back()[i].b;
                                            real_segment[i + 2].c = back()[i].c;
                                            real_segment[i + 2].k = back()[i].k;
                                        }
                                        real_segments.push_back(real_segment);
                                        real_frames_index.push_back(frame_count);
                                        frame_count++;
                                    }
                                    else
                                        // frame xy table bad_word!
                                        cout << "error on line: " << line_number << " \'" << word << "\' is not a LaserBoy text format token." << endl;
                                }
                                else
                                {
                                    // frame xy table short
                                    back().from_ifstream_txt(in, LASERBOY_2D_FRAME_TABLE, LASERBOY_SHORT, line_number);
                                    frame_count++;
                                }
                            }
                            else
                                // frame xy bad_word!
                                cout << "error on line: " << line_number << " \'" << word << "\' is not a LaserBoy text format token." << endl;
                        }
                        else
                        {
                            // frame xy rgb short
                            frame.from_ifstream_txt(in, LASERBOY_2D_FRAME_RGB, LASERBOY_SHORT, line_number);
                            push_back(frame);
                            frame_count++;
                        }
                    }
                    else if(word == "xyz") // 3D
                    {
                        // frame xyz
                        if(get_next_word(in, word, line_number))
                        {
                            if(word == "rgb")
                            {
                                // frame xyz rgb
                                if(get_next_word(in, word, line_number))
                                {
                                    if(word == "short")
                                    {
                                        // frame xyz rgb short
                                        frame.from_ifstream_txt(in, LASERBOY_3D_FRAME_RGB, LASERBOY_SHORT, line_number);
                                        push_back(frame);
                                        frame_count++;
                                    }
                                    else if(word == "unit")
                                    {
                                        // frame xyz rgb unit
                                        frame.from_ifstream_txt(in, LASERBOY_3D_FRAME_RGB, LASERBOY_UNIT, line_number);
                                        push_back(frame);
                                        frame_count++;
                                    }
                                    else if(word == "real")
                                    {
                                        // frame xyz rgb real
                                        real_segment.from_ifstream_txt(in, LASERBOY_3D_FRAME_RGB, line_number);
                                        real_segments.push_back(real_segment);
                                        real_frames_index.push_back(frame_count);
                                        frame.empty();
                                        push_back(frame);
                                        frame_count++;
                                    }
                                    else
                                        // frame xyz rgb bad_word!
                                        cout << "error on line: " << line_number << " \'" << word << "\' is not a LaserBoy text format token." << endl;
                                }
                                else
                                {
                                    // frame xyz rgb short
                                    frame.from_ifstream_txt(in, LASERBOY_3D_FRAME_RGB, LASERBOY_SHORT, line_number);
                                    push_back(frame);
                                    frame_count++;
                                }

                            }
                            else if(word == "hex")
                            {
                                // frame xyz hex
                                if(get_next_word(in, word, line_number))
                                {
                                    if(word == "short")
                                    {
                                        // frame xyz hex short
                                        frame.from_ifstream_txt(in, LASERBOY_3D_FRAME_HEX, LASERBOY_SHORT, line_number);
                                        push_back(frame);
                                        frame_count++;
                                    }
                                    else if(word == "unit")
                                    {
                                        // frame xyz hex unit
                                        frame.from_ifstream_txt(in, LASERBOY_3D_FRAME_HEX, LASERBOY_UNIT, line_number);
                                        push_back(frame);
                                        frame_count++;
                                    }
                                    else if(word == "real")
                                    {
                                        // frame xyz hex real
                                        real_segment.from_ifstream_txt(in, LASERBOY_3D_FRAME_HEX, line_number);
                                        real_segments.push_back(real_segment);
                                        real_frames_index.push_back(frame_count);
                                        frame.empty();
                                        push_back(frame);
                                        frame_count++;
                                    }
                                    else
                                        // frame xyz hex bad_word!
                                        cout << "error on line: " << line_number << " \'" << word << "\' is not a LaserBoy text format token." << endl;
                                }
                                else
                                {
                                    // frame xyz hex short
                                    frame.from_ifstream_txt(in, LASERBOY_3D_FRAME_HEX, LASERBOY_SHORT, line_number);
                                    push_back(frame);
                                    frame_count++;
                                }

                            }
                            else if(word == "palette")
                            {
                                // frame xyz palette
                                if(get_next_word(in, word, line_number))
                                {
                                    if(word == "short")
                                    {
                                        // frame xyz palette short
                                        frame.from_ifstream_txt(in, LASERBOY_3D_FRAME_PALETTE, LASERBOY_SHORT, line_number);
                                        push_back(frame);
                                        frame_count++;
                                    }
                                    else if(word == "unit")
                                    {
                                        // frame xyz palette unit
                                        frame.from_ifstream_txt(in, LASERBOY_3D_FRAME_PALETTE, LASERBOY_UNIT, line_number);
                                        push_back(frame);
                                        frame_count++;
                                    }
                                    else if(word == "real")
                                    {
                                        // frame xyz palette real
                                        real_segment.from_ifstream_txt(in, LASERBOY_3D_FRAME_PALETTE, line_number);
                                        real_segments.push_back(real_segment);
                                        real_frames_index.push_back(frame_count);
                                        frame.empty();
                                        push_back(frame);
                                        frame_count++;
                                    }
                                    else
                                        // frame xyz palette bad_word!
                                        cout << "error on line: " << line_number << " \'" << word << "\' is not a LaserBoy text format token." << endl;
                                }
                                else
                                {
                                    // frame xyz palette short
                                    frame.from_ifstream_txt(in, LASERBOY_3D_FRAME_PALETTE, LASERBOY_SHORT, line_number);
                                    push_back(frame);
                                    frame_count++;
                                }
                            }
                            else if(word == "table")
                            {
                                // frame xyz table
                                if(get_next_word(in, word, line_number))
                                {
                                    if(word == "short")
                                    {
                                        // frame xyz table short
                                        frame.from_ifstream_txt(in, LASERBOY_3D_FRAME_TABLE, LASERBOY_SHORT, line_number);
                                        push_back(frame);
                                        frame_count++;
                                    }
                                    else if(word == "unit")
                                    {
                                        // frame xyz table unit
                                        frame.from_ifstream_txt(in, LASERBOY_3D_FRAME_TABLE, LASERBOY_UNIT, line_number);
                                        push_back(frame);
                                        frame_count++;
                                    }
                                    else if(word == "real")
                                    {
                                        // frame xyz table real
                                        real_segment.from_ifstream_txt(in, LASERBOY_3D_FRAME_TABLE, line_number);
                                        for(int i = 0; i < back().number_of_vertices(); i++)
                                        {
                                            real_segment[i + 2].r = back()[i].r;
                                            real_segment[i + 2].g = back()[i].g;
                                            real_segment[i + 2].b = back()[i].b;
                                            real_segment[i + 2].c = back()[i].c;
                                            real_segment[i + 2].k = back()[i].k;
                                        }
                                        real_segments.push_back(real_segment);
                                        real_frames_index.push_back(frame_count);
                                        frame_count++;
                                    }
                                    else
                                        // frame xyz table bad_word!
                                        cout << "error on line: " << line_number << " \'" << word << "\' is not a LaserBoy text format token." << endl;
                                }
                                else
                                {
                                    // frame xyz table short
                                    frame.from_ifstream_txt(in, LASERBOY_3D_FRAME_TABLE, LASERBOY_SHORT, line_number);
                                    push_back(frame);
                                    frame_count++;
                                }
                            }
                            else
                                // frame xyz bad_word!
                                cout << "error on line: " << line_number << " \'" << word << "\' is not a LaserBoy text format token." << endl;
                        }
                        else
                        {
                            // frame xyz rgb short
                            frame.from_ifstream_txt(in, LASERBOY_3D_FRAME_RGB, LASERBOY_SHORT, line_number);
                            push_back(frame);
                            frame_count++;
                        }
                    }
                    else
                        // frame bad_word!
                        cout << "error on line: " << line_number << " \'" << word << "\' is not a LaserBoy text format token." << endl;
                }
                else
                {
                    // frame xy rgb short
                    frame.from_ifstream_txt(in, LASERBOY_2D_FRAME_RGB, LASERBOY_SHORT, line_number);
                    push_back(frame);
                    frame_count++;
                }
            }
            //----------------------------------------------------------------
            else if(word == "palette")
            {
                // palette
                if(get_next_word(in, word, line_number))
                {
                    if(word == "named")
                    {
                        if(get_next_word(in, word, line_number))
                        {
                            // palette named name
                            named_palette_index = p_space->name_to_palette_index(word);
                            if(named_palette_index == LASERBOY_NOT_FOUND)
                                cout << "error on line: " << line_number << " \'" << word << "\' is not a valid LaserBoy palette name." << endl;
                            else
                                p_space->palette_index = named_palette_index;
                        }
                        else
                            // palette named [no-name]
                            cout << "error on line: " << line_number << " LaserBoy palette name is missing." << endl;
                    }
                    else if(word == "rgb")
                    {
                        if(get_next_word(in, word, line_number))
                        {
                            // palette rgb name
                            named_palette_index = p_space->name_to_palette_index(word);
                            if(named_palette_index == LASERBOY_NOT_FOUND)
                                p_space->palette_from_ifstream_txt(in, LASERBOY_RGB, line_number, word);
                            else
                                p_space->palette_index = named_palette_index;
                        }
                        else
                            // palette rgb [no-name]
                            p_space->palette_from_ifstream_txt(in, LASERBOY_RGB, line_number);
                    }
                    else if(word == "hex")
                    {
                        if(get_next_word(in, word, line_number))
                        {
                            // palette hex name
                            named_palette_index = p_space->name_to_palette_index(word);
                            if(named_palette_index == LASERBOY_NOT_FOUND)
                                p_space->palette_from_ifstream_txt(in, LASERBOY_HEX, line_number, word);
                            else
                                p_space->palette_index = named_palette_index;
                        }
                        else
                            // palette hex [no-name]
                            p_space->palette_from_ifstream_txt(in, LASERBOY_HEX, line_number);
                    }
                    else
                        // palette bad_word!
                        cout << "error on line: " << line_number << " \'" << word << "\' is not a LaserBoy text format token." << endl;
                }
                else
                    // palette rgb [no-name]
                    p_space->palette_from_ifstream_txt(in, LASERBOY_RGB, line_number);
            }
            //----------------------------------------------------------------
            else if(word == "table")
            {
                if(get_next_word(in, word, line_number))
                {
                    if(word == "rgb")
                    {
                        // table rgb
                        frame.from_ifstream_txt(in, LASERBOY_TABLE, LASERBOY_RGB, line_number);
                        push_back(frame);
                    }
                    else if(word == "hex")
                    {
                        // table hex
                        frame.from_ifstream_txt(in, LASERBOY_TABLE, LASERBOY_HEX, line_number);
                        push_back(frame);
                    }
                    else
                        // table bad_word!
                        cout << "error on line: " << line_number << " \'" << word << "\' is not a LaserBoy text format token." << endl;
                }
                else
                {
                    // table rgb
                    frame.from_ifstream_txt(in, LASERBOY_TABLE, LASERBOY_RGB, line_number);
                    push_back(frame);
                }
            }
            else
                // bad_word!
                cout << "error on line: " << line_number << " \'" << word << "\' is not a LaserBoy text format token." << endl;
        } // end if(get_next_word(in, word, line_number))
    } // end while(in.good())
    //------------------------------------------------------------------------
    if(real_segments.size())
    {
        real_segments.normalize();
        for(i = 0; i < real_frames_index.size(); i++)
            at(real_frames_index[i]) = (LaserBoy_segment)real_segments[i];
    }
    //------------------------------------------------------------------------
    //if(!in.eof() && error && some_data_accepted)
    //    error = LASERBOY_OK;
    //------------------------------------------------------------------------
    return error;
}

//############################################################################
int LaserBoy_frame_set::from_ifstream_LaserBoy_wave(LaserBoy_wave_header header,
                                                    ifstream& in,
                                                    bool append
                                                   )
{
    int  i,
         span,
         last,
         red,
         green,
         blue,
         offset_max,
         offset_min;

    LaserBoy_frame   frame(p_space);
    LaserBoy_vertex  vertex        ;

    LaserBoy_wave_sample_pointer  roll_over    ;
    LaserBoy_wave_sample_pointer *sample_window;
    //------------------------------------------------------------------------
    offset_min = header.offset[0];
    offset_max = header.offset[0];
    for(i = 1; i < 6; i++)
    {
        if(header.offset[i] > offset_max)
            offset_max = header.offset[i];
        if(header.offset[i] < offset_min)
            offset_min = header.offset[i];
    }
    span  = (offset_max - offset_min) + 2; // size of sample_window
    last  = span - 1                     ; // last index of sample_window
    red   = -header.offset[2] + 1        ; // index of red
    green = -header.offset[3] + 1        ; // index of green
    blue  = -header.offset[4] + 1        ; // index of blue
    //------------------------------------------------------------------------
    sample_window = new LaserBoy_wave_sample_pointer[span];
    for(i = 0; i < span; i++)
        sample_window[i] = new LaserBoy_wave_sample(header.num_channels);
    //------------------------------------------------------------------------
    frame.format        = LASERBOY_2D_FRAME;
    frame.palette_index = LASERBOY_TRUE_COLOR;
    //------------------------------------------------------------------------
    if(!append) // to the current frame set
    {
        clear();
        num_2D_frames = 0;
        num_3D_frames = 0;
    }
    //------------------------------------------------------------------------
    while(true)
    {
        roll_over = sample_window[0];
        for(i = 0; i < last; i++)
            sample_window[i] = sample_window[i + 1];
        sample_window[last] = roll_over;

        if(!sample_window[last]->from_ifstream_wave(in)) // if EOF
            break;

        if((header.LaserBoy_wave_mode & LASERBOY_WAVE_POSITIVE) == 0)
            sample_window[last]->negate();
        //--------------------------------------------------------------------
        vertex.x = sample_window[1]->channel[0];
        vertex.y = sample_window[1]->channel[1];
        if(header.signal_id[5] == LASERBOY_SIGNAL_Z_POSITION)
            vertex.z = sample_window[1]->channel[5];
        else
            vertex.z = 0;
        //--------------------------------------------------------------------
        if(sample_window[red  ]->channel[2] > 0)
            vertex.r = (u_char)abs(sample_window[red  ]->channel[2] >> 7);
        else
            vertex.r = 0;
        //--------------------------------------------------------------------
        if(sample_window[green]->channel[3] > 0)
            vertex.g = (u_char)abs(sample_window[green]->channel[3] >> 7);
        else
            vertex.g = 0;
        //--------------------------------------------------------------------
        if(sample_window[blue ]->channel[4] > 0)
            vertex.b = (u_char)abs(sample_window[blue ]->channel[4] >> 7);
        else
            vertex.b = 0;
        //--------------------------------------------------------------------
        if(vertex.r + vertex.g + vertex.b)
            vertex.k = 0;
        else
            vertex.k = LASERBOY_BLANKING_BIT;
        vertex.c = 0;
        //--------------------------------------------------------------------
        frame.push_back(vertex);
        if(sample_window[red]->channel[2] & 0x0001) // end of frame
        {
            frame.is_unique = true;
            if(    (header.LaserBoy_wave_mode & LASERBOY_WAVE_UNIQUE_FRAME)  // if we are looking for it
                && (!(sample_window[green]->channel[3] & 0x0001)) // and we do not find it
              )
                frame.is_unique = false;
            frame.back().blank();
            frame.quantity = (u_short)frame.number_of_vertices();
            push_back(frame);
            frame.empty(); // clears frame and sets quantity to zero!
        }
    }
    //------------------------------------------------------------------------
    renumber();
    error = LASERBOY_OK;
    //------------------------------------------------------------------------
    delete sample_window;
    return error;
}

//############################################################################
int LaserBoy_frame_set::from_ifstream_unframed_wave(LaserBoy_wave_header header,
                                                    ifstream& in,
                                                    bool append
                                                   )
{
    int  i,
         vertices = 0,
         span,
         last,
         red,
         green,
         blue,
         offset_max,
         offset_min;

    LaserBoy_frame   frame(p_space);
    LaserBoy_vertex  vertex        ;

    LaserBoy_wave_sample_pointer  roll_over    ;
    LaserBoy_wave_sample_pointer *sample_window;
    //------------------------------------------------------------------------
    offset_min = header.offset[0];
    offset_max = header.offset[0];
    for(i = 1; i < 6; i++)
    {
        if(header.offset[i] > offset_max)
            offset_max = header.offset[i];
        if(header.offset[i] < offset_min)
            offset_min = header.offset[i];
    }
    span  = (offset_max - offset_min) + 2; // size of sample_window
    last  = span - 1                     ; // last index of sample_window
    red   = -header.offset[2] + 1        ; // index of red
    green = -header.offset[3] + 1        ; // index of green
    blue  = -header.offset[4] + 1        ; // index of blue
    //------------------------------------------------------------------------
    sample_window = new LaserBoy_wave_sample_pointer[span];
    for(i = 0; i < span; i++)
        sample_window[i] = new LaserBoy_wave_sample(header.num_channels);
    //------------------------------------------------------------------------
    frame.format        = LASERBOY_2D_FRAME;
    frame.palette_index = LASERBOY_TRUE_COLOR;
    //------------------------------------------------------------------------
    if(!append) // to the current frame set
    {
        clear();
        num_2D_frames = 0;
        num_3D_frames = 0;
    }
    //------------------------------------------------------------------------
    while(true)
    {
        roll_over = sample_window[0];
        for(i = 0; i < last; i++)
            sample_window[i] = sample_window[i + 1];
        sample_window[last] = roll_over;

        if(!sample_window[last]->from_ifstream_wave(in)) // if EOF
            break;

        if((header.LaserBoy_wave_mode & LASERBOY_WAVE_POSITIVE) == 0)
            sample_window[last]->negate();
        //--------------------------------------------------------------------
        vertex.x = sample_window[1]->channel[0];
        vertex.y = sample_window[1]->channel[1];
        if(header.signal_id[5] == LASERBOY_SIGNAL_Z_POSITION)
            vertex.z = sample_window[1]->channel[5];
        else
            vertex.z = 0;
        //--------------------------------------------------------------------
        if(sample_window[red  ]->channel[2] > 0)
            vertex.r = (u_char)abs(sample_window[red  ]->channel[2] >> 7);
        else
            vertex.r = 0;
        //--------------------------------------------------------------------
        if(sample_window[green]->channel[3] > 0)
            vertex.g = (u_char)abs(sample_window[green]->channel[3] >> 7);
        else
            vertex.g = 0;
        //--------------------------------------------------------------------
        if(sample_window[blue ]->channel[4] > 0)
            vertex.b = (u_char)abs(sample_window[blue ]->channel[4] >> 7);
        else
            vertex.b = 0;
        //--------------------------------------------------------------------
        if(vertex.r + vertex.g + vertex.b)
            vertex.k = 0;
        else
            vertex.k = LASERBOY_BLANKING_BIT;
        vertex.c = 0;
        //--------------------------------------------------------------------
        frame.push_back(vertex);
        vertices++;

        if(!(vertices % p_space->unframed_wave_stride))
        {
            frame.is_unique = true;
            frame.quantity = (u_short)frame.number_of_vertices();
            push_back(frame);
            frame.empty();
        }
    }
    //------------------------------------------------------------------------
    renumber();
    error = LASERBOY_OK;
    //------------------------------------------------------------------------
    delete sample_window;
    return error;
}

//############################################################################
int LaserBoy_frame_set::from_ifstream_unformatted_wave(  LaserBoy_wave_header header
                                                        ,ifstream&            in
                                                        , bool                append
                                                      )
{
    int  i,
         vertices = 0,
         span,
         last,
         red,
         green,
         blue,
         offset_max,
         offset_min;

    LaserBoy_frame   frame(p_space);
    LaserBoy_vertex  vertex;
    
    LaserBoy_wave_sample_pointer  roll_over;
    LaserBoy_wave_sample_pointer *sample_window;
    //------------------------------------------------------------------------
    offset_max = p_space->wave_offsets[0];
    for(i = 1; i < 6; i++)
        if(p_space->wave_offsets[i] > offset_max)
            offset_max = p_space->wave_offsets[i];
    //------------------------------------------------------------------------
    for(i = 0; i < 6; i++)
        p_space->wave_offsets[i] -= offset_max;
    //------------------------------------------------------------------------
    for(i = 0; i < 6; i++)
        header.offset[i] = p_space->wave_offsets[i]; // use current system offsets
    //------------------------------------------------------------------------
    offset_min = header.offset[0];
    offset_max = header.offset[0];
    for(i = 1; i < 6; i++)
    {
        if(header.offset[i] > offset_max)
            offset_max = header.offset[i];
        if(header.offset[i] < offset_min)
            offset_min = header.offset[i];
    }
    //------------------------------------------------------------------------
    span  = (offset_max - offset_min) + 2; // at least 2 samples!
    last  = span - 1;
    red   = -header.offset[2] + 1;
    green = -header.offset[3] + 1;
    blue  = -header.offset[4] + 1;
    //------------------------------------------------------------------------
    sample_window = new LaserBoy_wave_sample_pointer[span];
    for(i = 0; i < span; i++)
        sample_window[i] = new LaserBoy_wave_sample(header.num_channels);
    //------------------------------------------------------------------------
    frame.format        = LASERBOY_2D_FRAME;
    frame.palette_index = LASERBOY_TRUE_COLOR;
    //------------------------------------------------------------------------
    if(!append) // to the current frame set
    {
        clear();
        num_2D_frames = 0;
        num_3D_frames = 0;
    }
    //------------------------------------------------------------------------
    while(true)
    {
        roll_over = sample_window[0];
        for(i = 0; i < last; i++)
            sample_window[i] = sample_window[i + 1];
        sample_window[last] = roll_over;

        if(!sample_window[last]->from_ifstream_wave(in)) // if EOF
            break;

        vertices++;
        vertex.x = sample_window[1]->channel[0];
        vertex.y = sample_window[1]->channel[1];
        vertex.z = 0;
        switch(header.num_channels)
        {
            case 2: vertex.r = 255;
                    vertex.g = 255;
                    vertex.b = 255;
                    break;
            //----------------------------------------------------------------
            case 4: vertex.r =
                    vertex.g =
                    vertex.b = (u_char) abs(sample_window[red  ]->channel[2] / 128);
                    break;
            //----------------------------------------------------------------
            case 6:
            case 8: vertex.r = (u_char) abs(sample_window[red  ]->channel[2] / 128);
                    vertex.g = (u_char) abs(sample_window[green]->channel[3] / 128);
                    vertex.b = (u_char) abs(sample_window[blue ]->channel[4] / 128);
                    break;
            //----------------------------------------------------------------
        }
        vertex.k = 0;
        vertex.c = 0;
        //--------------------------------------------------------------------
        frame.push_back(vertex);

        if(!(vertices % p_space->unframed_wave_stride))
        {
            frame.is_unique = true;
            frame.quantity = (u_short)frame.number_of_vertices();
            push_back(frame);
            frame.empty();
        }
    }
    //------------------------------------------------------------------------
    frame.is_unique = true;
    frame.quantity = (u_short)frame.number_of_vertices();
    push_back(frame);
    frame.empty(); // clears frame and sets quantity to zero!
    renumber();
    error = LASERBOY_OK;
    return error;
}

//############################################################################
bool LaserBoy_frame_set::is_2D()
{
    for(int i = 0; i < number_of_frames(); i++)
        if(!at(i).is_2D())
            return false;
    return true;
}

//############################################################################
void LaserBoy_frame_set::flatten_z()
{
    for(int i = 0; i < number_of_frames(); i++)
        at(i).flatten_z();
    return;
}

//############################################################################
void LaserBoy_frame_set::z_order_vertices(unsigned short span)
{
    for(int i = 0; i < number_of_frames(); i++)
        at(i).z_order_vertices(span);
    return;
}

//############################################################################
void LaserBoy_frame_set::unblank_all_vertices()
{
    for(int i = 0; i < number_of_frames(); i++)
        at(i).unblank_all_vertices();
    return;
}

//############################################################################
void LaserBoy_frame_set::convert_black_to_blank()
{
    for(int i = 0; i < number_of_frames(); i++)
        at(i).convert_black_to_blank();
    return;
}

//############################################################################
void LaserBoy_frame_set::convert_blank_to_black()
{
    for(int i = 0; i < number_of_frames(); i++)
        at(i).convert_blank_to_black();
    return;
}

//############################################################################
void LaserBoy_frame_set::reduce_blank_vectors()
{
    for(int i = 0; i < number_of_frames(); i++)
        at(i).reduce_blank_vectors();
    return;
}

//############################################################################
void LaserBoy_frame_set::remove_dots()
{
    for(int i = 0; i < number_of_frames(); i++)
        at(i).remove_dots();
    return;
}

//############################################################################
void LaserBoy_frame_set::enhance_dots()
{
    for(int i = 0; i < number_of_frames(); i++)
        at(i).enhance_dots();
    return;
}

//############################################################################
void LaserBoy_frame_set::remove_dwell_vertices()
{
    for(int i = 0; i < number_of_frames(); i++)
        at(i).remove_dwell_vertices();
    return;
}

//############################################################################
void LaserBoy_frame_set::reduce_lit_vectors()
{
    for(int i = 0; i < number_of_frames(); i++)
        at(i).reduce_lit_vectors();
    return;
}

//############################################################################
void LaserBoy_frame_set::minimize()
{
    for(int i = 0; i < number_of_frames(); i++)
        at(i).minimize();
    return;
}

//############################################################################
void LaserBoy_frame_set::to_dots()
{
    for(int i = 0; i < number_of_frames(); i++)
        at(i).to_dots();
    return;
}

//############################################################################
void LaserBoy_frame_set::accentuate_the_positive()
{
    for(int i = 0; i < number_of_frames(); i++)
        at(i).accentuate_the_positive();
    return;
}

//############################################################################
void LaserBoy_frame_set::eliminate_the_negative()
{
    for(int i = 0; i < number_of_frames(); i++)
        at(i).eliminate_the_negative();
    return;
}

//############################################################################
void LaserBoy_frame_set::impose_bit_resolution()
{
    for(int i = 0; i < number_of_frames(); i++)
        at(i).impose_bit_resolution();
    return;
}

//############################################################################
void LaserBoy_frame_set::explode_current_frame()
{
    int                 i;
    LaserBoy_frame_set  results(p_space),
                        exploded = current_frame().find_segments();

    for(i = 0; i < frame_index; i++)
        results += at(i);

    results += exploded;

    for(i = frame_index + 1; i < number_of_frames(); i++)
        results += at(i);

    *this = results;
    renumber();
    return;
}

//############################################################################
void LaserBoy_frame_set::explode_all_frames()
{
    int                 i;
    LaserBoy_frame_set  results (p_space),
                        exploded(p_space);

    for(i = 0; i < number_of_frames(); i++)
        results +=  at(i).find_segments();

    *this = results;
    renumber();
    return;
}

//############################################################################
void LaserBoy_frame_set::collapse_selected_frames()
{
    LaserBoy_frame  collapsed(p_space);
    int             first_selected_frame_index = -1;

    for(int i = 0; i < number_of_frames(); i++)
    {
        if(at(i).is_selected)
        {
            if(first_selected_frame_index == -1)
            {
                first_selected_frame_index = i;
                collapsed.palette_index = at(i).palette_index;
            }
            collapsed += at(i);
        }
    }
    if(first_selected_frame_index != -1)
    {
        frame(first_selected_frame_index) = collapsed;
        frame(first_selected_frame_index).is_selected = false;
        delete_selected_frames();
        renumber();
    }
    return;
}

//############################################################################
void LaserBoy_frame_set::delete_every_nth_frame(unsigned short n)
{
    LaserBoy_frame_set results(p_space);
    if(n > 1 && n < number_of_frames())
    {
        for(int i = 0; i < number_of_frames(); i++)
            if(i % n)
                results += at(i);
        *this = results;
        renumber();
    }
    return;
}

//############################################################################
void LaserBoy_frame_set::fracture_segments()
{
    for(int i = 0; i < number_of_frames(); i++)
        at(i).fracture_segments();
    return;
}

//############################################################################
void LaserBoy_frame_set::blank_equivalent_vectors()
{
    for(int i = 0; i < number_of_frames(); i++)
        at(i).blank_equivalent_vectors();
    return;
}

//############################################################################
void LaserBoy_frame_set::bond_segments()
{
    for(int i = 0; i < number_of_frames(); i++)
        at(i).bond_segments();
    return;
}

//############################################################################
void LaserBoy_frame_set::reorder_segments()
{
    for(int i = 0; i < number_of_frames(); i++)
        at(i).reorder_segments();
    return;
}

//############################################################################
void LaserBoy_frame_set::add_dwell_vertices()
{
    for(int i = 0; i < number_of_frames(); i++)
        at(i).add_dwell_vertices();
    return;
}

//############################################################################
void LaserBoy_frame_set::add_lit_span_vertices()
{
    for(int i = 0; i < number_of_frames(); i++)
        at(i).add_lit_span_vertices();
    return;
}

//############################################################################
void LaserBoy_frame_set::add_blank_span_vertices()
{
    for(int i = 0; i < number_of_frames(); i++)
        at(i).add_blank_span_vertices();
    return;
}

//############################################################################
void LaserBoy_frame_set::delete_redundant_frames()
{
    LaserBoy_frame_set reduced(p_space);
    for(int i = 0; i < number_of_frames(); i++)
        if(at(i).is_unique)
            reduced.push_back(at(i));
    reduced.renumber();
    *this = reduced;
    return;
}

//############################################################################
void LaserBoy_frame_set::optimize()
{
    if(number_of_frames())
    {
        int i,
            dwell_vertex_count = 0,
            dwell_vertex_total = 0,
            blank_dwell        = p_space->sample_rate / 220;
        //--------------------------------------------------------------------
        for(i = 0; i < number_of_frames(); i++)
            at(i).optimize();
        //--------------------------------------------------------------------
        intro.clear();
        for(i = 0; i < blank_dwell; i++)
            intro.push_back(LaserBoy_vertex()); // origin
        //--------------------------------------------------------------------
        intro += LaserBoy_segment(p_space, intro[0], front()[0]); // from origin to first vertex in first frame
        //--------------------------------------------------------------------
        for(i = 0; i < blank_dwell; i++)
            intro.push_back(front()[0]);
        //--------------------------------------------------------------------
        for(i = 0; i < (number_of_frames() - 1); i++)
        {
            at(i).transition_to_next_frame.empty(); // clears frame and sets quantity to zero!
            dwell_vertex_total = dwell_vertices(delta_angle(at(i)[at(i).number_of_vertices() - 2] | at(i + 1).front(),
                                                           at(i)[at(i).number_of_vertices() - 2] | at(i).back(),
                                                           at(i).back()                        | at(i + 1).front()
                                                          ),
                                               p_space->sample_rate,
                                               p_space->max_dwell_microsec
                                              );
            if(dwell_vertex_total)
                for(dwell_vertex_count = 0; dwell_vertex_count < dwell_vertex_total; dwell_vertex_count++)
                    at(i).transition_to_next_frame.push_back(at(i).back());
            //----------------------------------------------------------------
            at(i).transition_to_next_frame += LaserBoy_segment(p_space, at(i).back(), at(i + 1)[0]);
        }
        //--------------------------------------------------------------------
        back().transition_to_next_frame = LaserBoy_segment(p_space, back().back(), intro[0]);
        for(i = 0; i < blank_dwell; i++)
            back().transition_to_next_frame.push_back(intro[0]);
        //--------------------------------------------------------------------
    }
    return;
}

//############################################################################
void LaserBoy_frame_set::rainbow_recolor(int effect)
{
    for(int i = 0; i < number_of_frames(); i++)
        at(i).rainbow_recolor(effect);
    return;
}

//############################################################################
void LaserBoy_frame_set::quarter_turn(int plane, int turns)
{
    for(int i = 0; i < number_of_frames(); i++)
        at(i).quarter_turn(plane, turns);
    return;
}

//############################################################################
int LaserBoy_frame_set::scale_around_origin(LaserBoy_3D_double factor)
{
    int all_in_bounds = true;
    for(int i = 0; i < number_of_frames(); i++)
        all_in_bounds &= at(i).scale_around_origin(factor);
    return all_in_bounds;
}

//############################################################################
int LaserBoy_frame_set::scale_around_origin()
{
    int all_in_bounds = true;
    for(int i = 0; i < number_of_frames(); i++)
        all_in_bounds &= at(i).scale_around_origin();
    return all_in_bounds;
}

//############################################################################
int LaserBoy_frame_set::rotate_around_origin(LaserBoy_3D_double angle)
{
    int all_in_bounds = true;
    for(int i = 0; i < number_of_frames(); i++)
        all_in_bounds &= at(i).rotate_around_origin(angle);
    return all_in_bounds;
}

//############################################################################
int LaserBoy_frame_set::rotate_around_origin()
{
    int all_in_bounds = true;
    for(int i = 0; i < number_of_frames(); i++)
        all_in_bounds &= at(i).rotate_around_origin();
    return all_in_bounds;
}

//############################################################################
int LaserBoy_frame_set::move(LaserBoy_3D_double offset)
{
    int all_in_bounds = true;
    for(int i = 0; i < number_of_frames(); i++)
        all_in_bounds &= at(i).move(offset);
    return all_in_bounds;
}

//############################################################################
int LaserBoy_frame_set::move()
{
    int all_in_bounds = true;
    for(int i = 0; i < number_of_frames(); i++)
        all_in_bounds &= at(i).move();
    return all_in_bounds;
}

//############################################################################
void LaserBoy_frame_set::toggle_frame_select_start()
{
    int i;
    if(frame_select_start == -1) // was unset
    {
        frame_select_start = frame_index; // set
        if(frame_select_end != -1)
        {
            for(i = frame_select_start;
                i != frame_select_end;
                (frame_select_start <= frame_select_end) ? (i++) : (i--)
               )
                at(i).is_selected = true;
            at(frame_select_end).is_selected = true;
        }
    }
    else // was set
    {
        if(frame_select_end != -1) // was set
        {
            for(i = frame_select_start;
                i != frame_select_end;
                (frame_select_start <= frame_select_end) ? (i++) : (i--)
               )
                at(i).is_selected = false;
            at(frame_select_end).is_selected = false;
        }
        frame_select_start = -1; // unset
    }
    return;
}

//############################################################################
void LaserBoy_frame_set::toggle_frame_select_end()
{
    int i;
    if(frame_select_end == -1) // was unset
    {
        frame_select_end = frame_index; // set
        if(frame_select_start != -1)
        {
            for(i = frame_select_start;
                i != frame_select_end;
                (frame_select_start <= frame_select_end) ? (i++) : (i--)
               )
                at(i).is_selected = true;
            at(frame_select_end).is_selected = true;
        }
    }
    else // was set
    {
        if(frame_select_start != -1) // was set
        {
            for(i = frame_select_start;
                i != frame_select_end;
                (frame_select_start <= frame_select_end) ? (i++) : (i--)
               )
                at(i).is_selected = false;
            at(frame_select_end).is_selected = false;
        }
        frame_select_end = -1; // unset
    }
    return;
}

//############################################################################
void LaserBoy_frame_set::delete_selected_frames()
{
    LaserBoy_frame_set reduced(p_space);
    for(int i = 0; i < number_of_frames(); i++)
    {
        if(!at(i).is_selected)
            reduced.push_back(at(i));
    }
    reduced.renumber();

    if(    frame_select_start != -1
        && frame_select_end   != -1
      )
    {
        frame_index = (frame_select_start - 1);
        frame_select_start = -1;
        frame_select_end   = -1;
    }

    if(frame_index < 0)
        frame_index = 0;

    if(frame_index > reduced.number_of_frames())
        frame_index = reduced.number_of_frames() - 1;

    *this = reduced;
    return;
}

//############################################################################
void LaserBoy_frame_set::trim_to_selected_frames()
{
    LaserBoy_frame_set reduced(p_space);
    reduced.reserve(size());
    for(int i = 0; i < number_of_frames(); i++)
    {
        if(at(i).is_selected)
            reduced.push_back(at(i));
    }
    reduced.renumber();
    if(    frame_select_start != -1
        && frame_select_end   != -1
      )
    {
        frame_select_start = -1;
        frame_select_end   = -1;
    }
    frame_index = 0;
    *this = reduced;
    return;
}

//############################################################################
void LaserBoy_frame_set::bit_reduce_to_palette()
{
    for(int i = 0; i < number_of_frames(); i++)
        at(i).bit_reduce_to_palette();
    return;
}

//############################################################################
void LaserBoy_frame_set::best_reduce_to_palette()
{
    for(int i = 0; i < number_of_frames(); i++)
        at(i).best_reduce_to_palette();
    return;
}

//############################################################################
void LaserBoy_frame_set::promote_to_true_color()
{
    for(int i = 0; i < number_of_frames(); i++)
        at(i).promote_to_true_color();
    return;
}

//############################################################################
void LaserBoy_frame_set::best_match_palette(int index)
{
    if(!p_space->allow_lit_black)
        convert_black_to_blank();
    for(int i = 0; i < number_of_frames(); i++)
        at(i).best_match_palette(index);
    return;
}

//############################################################################
void LaserBoy_frame_set::best_match_target_palette()
{
    if(!p_space->allow_lit_black)
        convert_black_to_blank();
    for(int i = 0; i < number_of_frames(); i++)
        at(i).best_match_target_palette();
    return;
}

//############################################################################
void LaserBoy_frame_set::strip_color()
{
    for(int i = 0; i < number_of_frames(); i++)
        at(i).strip_color();
    return;
}

//############################################################################
void LaserBoy_frame_set::strip_color_or()
{
    for(int i = 0; i < number_of_frames(); i++)
        at(i).strip_color_or();
    return;
}

//############################################################################
void LaserBoy_frame_set::strip_color_avg()
{
    for(int i = 0; i < number_of_frames(); i++)
        at(i).strip_color_avg();
    return;
}

//############################################################################
void LaserBoy_frame_set::accept_palette(int index)
{
    for(int i = 0; i < number_of_frames(); i++)
        at(i).accept_palette(index);
    return;
}

//############################################################################
void LaserBoy_frame_set::accept_target_palette()
{
    for(int i = 0; i < number_of_frames(); i++)
        at(i).accept_target_palette();
    return;
}

//############################################################################
bool LaserBoy_frame_set::color_from_bmp(string file)
{
    char file_name[256];
    struct LaserBoy_bmp bmp;

    strcpy(file_name, file.c_str());
    if(bmp_from_file(&bmp, file_name))
    {
        color_from_bmp(&bmp);
        bmp_free(&bmp);
        return true;
    }
    return false;
}

//############################################################################
void LaserBoy_frame_set::color_from_bmp(struct LaserBoy_bmp* bmp)
{
    for(int i = 0; i < number_of_frames(); i++)
        at(i).color_from_bmp(bmp);
    return;
}

//############################################################################
bool LaserBoy_frame_set::subtract_bmp(string file)
{
    char file_name[81];
    struct LaserBoy_bmp bmp;

    strcpy(file_name, (file).c_str());
    if(bmp_from_file(&bmp, file_name))
    {
        subtract_bmp(&bmp);
        bmp_free(&bmp);
        return true;
    }
    return false;
}

//############################################################################
void LaserBoy_frame_set::subtract_bmp(struct LaserBoy_bmp* bmp)
{
    for(int i = 0; i < number_of_frames(); i++)
        at(i).subtract_bmp(bmp);
    return;
}

//############################################################################
void LaserBoy_frame_set::flip(int plane)
{
    for(int i = 0; i < number_of_frames(); i++)
        at(i).flip(plane);
    return;
}

//############################################################################
void LaserBoy_frame_set::renumber()
{
    int   i,
          local_palette_index = LASERBOY_ILDA_DEFAULT,
          added_palettes      = 0,
          frame_id            = 0;
    bool  in_true_color       = false;
    char  number[7];
    //------------------------------------------------------------------------
    for(i = 0; i < number_of_frames(); i++)
    {
        if((at(i).quantity = at(i).number_of_vertices()) > 1)
        {
            if(!p_space->dumb_file_format)
            {
                if(at(i).palette_index == LASERBOY_TRUE_COLOR)
                {
                    if(!in_true_color)
                    {
                        frame_id++;
                        added_palettes++; // add LASERBOY_332
                        local_palette_index = LASERBOY_LASERBOY_332;
                        in_true_color = true;
                    }
                    frame_id++;
                    added_palettes++; // add true_color
                }
                else if(at(i).palette_index != local_palette_index)
                {
                    frame_id++;
                    added_palettes++;
                    local_palette_index = at(i).palette_index;
                    in_true_color = false;
                }
            }
            sprintf(number, "%06d", frame_id);
            at(i).identity = frame_id;
            at(i).name     = "ES" + string(number);
            frame_id++;
        }
        else
            delete_frame(i);
    }
    //------------------------------------------------------------------------
    for(i = 0; i < number_of_frames(); i++)
        at(i).total = (u_short)(number_of_frames() + added_palettes);
    if(frame_index >= number_of_frames())
        frame_index = number_of_frames() - 1;
    return;
}

//############################################################################
LaserBoy_frame_set& LaserBoy_frame_set::reverse()
{
    LaserBoy_frame_set reversed(p_space);
    for(int i = 1; i <= number_of_frames(); i++)
        reversed.push_back(at(number_of_frames() - i));
    clear();
    insert(begin(), reversed.begin(), reversed.end());
    renumber();
    return *this;
}

//############################################################################
LaserBoy_frame_set& LaserBoy_frame_set::make_current_first()
{
    int                i,
                       j;
    LaserBoy_frame_set new_order(p_space);
    //------------------------------------------------------------------------
    for(i = frame_index; i < number_of_frames(); i++)
        new_order += at(i);
    //------------------------------------------------------------------------
    for(i = 0; i < frame_index; i++)
        new_order += at(i);
    //------------------------------------------------------------------------
    *this = new_order;
    frame_index = 0;
    //------------------------------------------------------------------------
    return *this;
}

//############################################################################
LaserBoy_frame_set& LaserBoy_frame_set::reverse_selected()
{
    int                i,
                       j;
    LaserBoy_frame_set selected(p_space);
    //------------------------------------------------------------------------
    for(i = 0; i < number_of_frames(); i++)
        if(at(i).is_selected)
            selected += at(i);
    //------------------------------------------------------------------------
    j = selected.number_of_frames();
    //------------------------------------------------------------------------
    for(i = 0; i < number_of_frames(); i++)
        if(at(i).is_selected)
            at(i) = selected[--j];
    //------------------------------------------------------------------------
    return *this;
}

//############################################################################
LaserBoy_frame_set& LaserBoy_frame_set::random_order()
{
    LaserBoy_frame_set random(p_space);
    vector<bool>       been_here(number_of_frames());
    //------------------------------------------------------------------------
    int random_index = rand() % number_of_frames();
    //------------------------------------------------------------------------
    for(int i = 0; i < number_of_frames(); i++)
        been_here[i] = false;
    //------------------------------------------------------------------------
    for(int i = 0; i < number_of_frames(); i++)
    {
        while(been_here[random_index])
        {
            random_index = rand() % number_of_frames();
        }
        random.push_back(at(random_index));
        been_here[random_index] = true;
    }
    //------------------------------------------------------------------------
    clear();
    insert(begin(), random.begin(), random.end());
    renumber();
    return *this;
}

//############################################################################
LaserBoy_frame_set& LaserBoy_frame_set::frame_reverse()
{
    for(int i = 1; i <= number_of_frames(); i++)
        at(i).reverse();
    return *this;
}

//############################################################################
bool LaserBoy_frame_set::save_as_ild(string file)
{
    ofstream out(file.c_str(), ios::out | ios::binary);
    if(out.is_open())
    {
        to_ofstream_ild(out);
        out.close();
    }
    else
        return false;
    return true;
}

//############################################################################
bool LaserBoy_frame_set::save_as_txt(string file)
{
    ofstream out(file.c_str(), ios::out);
    if(out.is_open())
    {
        txt_tag(out);
        to_ofstream_txt(out);
        out.close();
    }
    else
        return false;
    return true;
}

//############################################################################
void LaserBoy_frame_set::to_ofstream_ild(ofstream& out)
{
    int              file_section_number = 0,
                     local_palette_index = LASERBOY_ILDA_DEFAULT;
    bool             in_true_color       = false;
    LaserBoy_header  header;
    //------------------------------------------------------------------------
    renumber();
    header.format  = LASERBOY_PALETTE;
    header.total   = front().total;
    //------------------------------------------------------------------------
    for(int i = 0; i < number_of_frames(); i++)
    {
        if(!p_space->dumb_file_format)
        {
            if(at(i).palette_index == LASERBOY_TRUE_COLOR)
            {
                if(!in_true_color)
                {
                    header.name         = p_space->palette_picker(LASERBOY_LASERBOY_332).name;
                    header.quantity     = p_space->palette_picker(LASERBOY_LASERBOY_332).number_of_colors();
                    header.identity     = file_section_number;
                    p_space->palette_picker(LASERBOY_LASERBOY_332).to_ofstream_ild(out, header);
                    local_palette_index = LASERBOY_LASERBOY_332;
                    file_section_number++;
                    in_true_color = true;
                }
                file_section_number++;
            }
            else if(at(i).palette_index != local_palette_index)
            {
                header.name         = p_space->palette_picker(at(i).palette_index).name;
                header.quantity     = p_space->palette_picker(at(i).palette_index).number_of_colors();
                header.identity     = file_section_number;
                p_space->palette_picker(at(i).palette_index).to_ofstream_ild(out, header);
                local_palette_index = at(i).palette_index;
                file_section_number++;
                in_true_color = false;
            }
        }
        at(i).is_2D(); // checks for z and sets format accordingly;
        at(i).to_ofstream_ild(out);
        file_section_number++;
    }
    //------------------------------------------------------------------------
    header.format   = LASERBOY_3D_FRAME  ;
    header.name     = "ILDA_End"         ;
    header.quantity = 0                  ;
    header.identity = file_section_number;
    header.to_ofstream_ild(out)          ;
    return;
}

//############################################################################
void LaserBoy_frame_set::to_ofstream_txt(ofstream& out)
{
    int  local_palette_index = LASERBOY_TRUE_COLOR;
    for(int i = 0; i < number_of_frames(); i++)
    {
        if(!p_space->save_txt_with_color)
        {
            if(    at(i).palette_index != local_palette_index
                && at(i).palette_index != LASERBOY_TRUE_COLOR
              )
            {
                if(    p_space->save_txt_named_palettes
                    || at(i).palette_index >= LASERBOY_BUILT_IN_PALETTES
                  )
                    p_space->palette_picker(at(i).palette_index).to_ofstream_txt(out);
                else
                {
                    out << "palette named "
                        << p_space->palette_picker(at(i).palette_index).name
                        << endl
                        << endl;
                }
                local_palette_index = at(i).palette_index;
            }
            if(at(i).palette_index == LASERBOY_TRUE_COLOR)
                at(i).to_ofstream_txt_table(out);
        }
        //--------------------------------------------------------------------
        at(i).to_ofstream_txt(out);
    }
    //------------------------------------------------------------------------
    return;
}

//############################################################################
bool LaserBoy_frame_set::save_as_wave(string file)
{
    ofstream out(file.c_str(), ios::out | ios::binary);
    if(out.is_open())
    {
        if(number_of_frames() > 1)
            to_ofstream_wave(out);
        else
        {
            LaserBoy_frame copy(front());
            copy.rotate_around_origin();
            if(p_space->auto_flatten_z)
                copy.flatten_z();
            if(p_space->auto_optimize)
                copy.optimize();
            copy.save_as_wave(file);
        }
        out.close();
        return true;
    }
    return false;
}

//############################################################################
void LaserBoy_frame_set::to_ofstream_wave(ofstream& out)
{
    int                   i;
    LaserBoy_wave_header  header(   p_space->sample_rate
                                  ,    LASERBOY_WAVE_POSITIVE
                                     | LASERBOY_WAVE_END_OF_FRAME
                                     | LASERBOY_WAVE_UNIQUE_FRAME
                                     | LASERBOY_WAVE_SIGNAL_MATRIX
                                  , 6
                                );
    //------------------------------------------------------------------------
    LaserBoy_frame_set  copy(*this);
    //------------------------------------------------------------------------
    header.signal_id [0] = LASERBOY_SIGNAL_X_POSITION  ; // default values
    header.signal_id [1] = LASERBOY_SIGNAL_Y_POSITION  ;
    header.signal_id [2] = LASERBOY_SIGNAL_RED_ANALOG  ;
    header.signal_id [3] = LASERBOY_SIGNAL_GREEN_ANALOG;
    header.signal_id [4] = LASERBOY_SIGNAL_BLUE_ANALOG ;
    header.signal_id [5] = LASERBOY_SIGNAL_Z_POSITION  ;

    header.LSB_tag   [0] = LASERBOY_LSB_NOT_USED    ; // default values
    header.LSB_tag   [1] = LASERBOY_LSB_NOT_USED    ;
    header.LSB_tag   [2] = LASERBOY_LSB_END_OF_FRAME;
    header.LSB_tag   [3] = LASERBOY_LSB_UNIQUE_FRAME;
    header.LSB_tag   [4] = LASERBOY_LSB_NOT_USED    ;
    header.LSB_tag   [5] = LASERBOY_LSB_NOT_USED    ;

    if(    p_space->signal_bit_mask[0] != 0
        || p_space->signal_bit_mask[1] != 0
        || p_space->signal_bit_mask[2]  > 7
        || p_space->signal_bit_mask[3]  > 7
        || p_space->signal_bit_mask[4]  > 7
        || p_space->signal_bit_mask[5] != 0
      )
    {
        header.resolution[0] = 16 - p_space->signal_bit_mask[0];
        header.resolution[1] = 16 - p_space->signal_bit_mask[1];
        header.resolution[2] = 16 - p_space->signal_bit_mask[2];
        header.resolution[3] = 16 - p_space->signal_bit_mask[3];
        header.resolution[4] = 16 - p_space->signal_bit_mask[4];
        header.resolution[5] = 16 - p_space->signal_bit_mask[5];
        header.LaserBoy_wave_mode |= LASERBOY_SIGNAL_BIT_RESOLUTION;
    }
    //------------------------------------------------------------------------
    for(i = 0; i < copy.number_of_frames(); i++)
        copy[i].rotate_around_origin();
    //------------------------------------------------------------------------
    if(is_2D() || p_space->auto_flatten_z)
    {
        header.signal_id [5] = p_space->signal_if_not_z;
        header.resolution[5] = 8;
        copy.flatten_z();
    }
    //------------------------------------------------------------------------
    if(p_space->auto_optimize)
    {
        header.LaserBoy_wave_mode |= LASERBOY_WAVE_OPTIMIZED;
        header.parms.lit_dwell_overhang     =        p_space->lit_dwell_overhang    ;
        header.parms.lit_delta_max          = (float)p_space->lit_delta_max         ;
        header.parms.blank_delta_max        = (float)p_space->blank_delta_max       ;
        header.parms.max_dwell_microsec     = (float)p_space->max_dwell_microsec    ;
        header.parms.insignificant_distance = (float)p_space->insignificant_distance;
        header.parms.insignificant_angle    = (float)p_space->insignificant_angle   ;
        header.parms.frames_per_second      = (float)p_space->frames_per_second     ;
        copy.optimize();
    }
    //------------------------------------------------------------------------
    if(p_space->invert_wave_output)
        header.LaserBoy_wave_mode &= ~LASERBOY_WAVE_POSITIVE;
    //------------------------------------------------------------------------
    header.to_ofstream_wave(out); // take up the space!
    //------------------------------------------------------------------------
    if(p_space->auto_optimize)
        copy.intro.to_ofstream_wave(out, header);
    //------------------------------------------------------------------------
    for(i = 0; i < copy.number_of_frames(); i++)
        copy.at(i).to_ofstream_wave(out, header);
    //------------------------------------------------------------------------
    header.to_ofstream_wave(out);
    //------------------------------------------------------------------------
    return;
}

//############################################################################
bool LaserBoy_frame_set::save_as_bmp_directory(string dir) const
{
    char name[8];
#ifndef WIN32
    mkdir(dir.c_str(), 0700);
#else
    mkdir(dir.c_str());
#endif
    for(int i = 0; i < number_of_frames(); i++)
    {
        sprintf(name, "%06d", i);
        at(i).save_as_bmp(string(dir + '/' + name));
    }
    return true;
}

//############################################################################
bool LaserBoy_frame_set::save_as_dxf_directory(string dir) const
{
    bool all_saved = true;
    char name[8];
#ifndef WIN32
    mkdir(dir.c_str(), 0700);
#else
    mkdir(dir.c_str());
#endif
    for(int i = 0; i < number_of_frames(); i++)
    {
        sprintf(name, "%06d", i);
        all_saved &= at(i).save_as_dxf(string(dir + '/' + name + ".dxf"));
    }
    return all_saved;
}

//############################################################################
bool LaserBoy_frame_set::save_as_code(string object_name, string dir)
{
    ofstream out((dir + object_name + ".cpp").c_str(), ios::out);
    if(out.is_open())
    {
        out << "//############################################################################\n"
               "// ~  This file is the product of LaserBoy!  ~\n"
               "//\n"
               "// by James Lehman\n"
               "// james@akrobiz.com\n"
               "//\n"
               "//############################################################################\n\n\n\n"
               "//############################################################################\n"
            << "LaserBoy_frame_set ";
        out << object_name;
        out << "()\n"
               "{\n"
               "    LaserBoy_frame     frame;\n"
               "    LaserBoy_frame_set frames;\n"
               "    int                frame_index;\n\n"
               "    for(frame_index = 0; frame_index < ";
        out << number_of_frames();
        out << "; frame_index++)\n"
               "    {\n";
        for(int i = 0; i < number_of_frames(); i++)
        {

            at(i).to_ofstream_code(out, "        ");
            out << "        frames.push_back(frame);\n"
                   "        frame.empty();\n\n";
        }
        out << "    }\n"
               "    return frames;\n"
               "}\n\n"
               "//############################################################################\n"
               "//////////////////////////////////////////////////////////////////////////////\n"
               "//############################################################################\n\n\n";

        out.close();
        return true;
    }
    return false;
}

//############################################################################
//////////////////////////////////////////////////////////////////////////////
//############################################################################
